package com.sun.xacml.finder.impl;

import java.net.URI;
import java.net.URISyntaxException;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import com.sun.xacml.EvaluationCtx;
import com.sun.xacml.attr.AttributeDesignator;
import com.sun.xacml.attr.AttributeValue;
import com.sun.xacml.attr.BagAttribute;
import com.sun.xacml.attr.StringAttribute;
import com.sun.xacml.cond.EvaluationResult;
import com.sun.xacml.finder.AttributeFinderModule;

/**
 * 
 * @author romain.ferrari[AT]thalesgroup.com
 * @version 1.0
 * 
 *          Class implementing an Attribute finder for database. Arguments
 *          needed from the config.xml file :
 *          <url>jdbc:mysql://localhost:3305/</url>
 *			<username>root</username>
 *			<password></password>
 *			<dbName>keystone-rfo</dbName>
 *			<driver>com.mysql.jdbc.Driver</driver>
 *			<attributeSupportedId>tenantMember</attributeSupportedId>
 *			<sqlRequest>SELECT u.name as $alias 
 *						FROM user_roles ur, tenants t, users u 
 *						WHERE ur.tenant_id = t.id 
 *						AND u.id = ur.user_id 
 *						AND t.id = $filter
 *			</sqlRequest>
 *			<substituteValue>urn:oasis:names:tc:xacml:1.0:resource:resource-id</substituteValue>
 * 
 *          While $filter is the variable part used to make a filter in the SQL
 *          query mapped to the AttributeValue in the XACML request defined with
 *          the substituteValue option.
 * 
 *          $alias is used to make the developper's life easier and with the
 *          parsing of the SQL response.
 * 
 *          The column name where we extract the values is defined by the
 *          attributeSupportedId option.
 */
public class AttributeDBFinder extends AttributeFinderModule {

	private static final org.apache.log4j.Logger LOGGER = org.apache.log4j.Logger
			.getLogger(AttributeDBFinder.class);

	private String username;
	private String password;
	private String dbName;
	private String url;
	private String substituteValue;
	private String attributeSupportedId;
	private String sqlRequest;
	private String driver;

	private URI resourceId = null;
	private URI resourceType = null;
	private URI resourceCategoryUri = null;

	private StringAttribute resourceVal;

	public AttributeDBFinder() {
	}

	public AttributeDBFinder(List args) {
		Iterator it = args.iterator();
		while (it.hasNext()) {
			Object val = it.next();
			if (val instanceof HashMap) {
				HashMap map = (HashMap) val;
				if (map.containsKey("url")) {
					url = (String) map.get("url");
				} else if (map.containsKey("dbName")) {
					dbName = (String) map.get("dbName");
				} else if (map.containsKey("driver")) {
					driver = (String) map.get("driver");
				} else if (map.containsKey("sqlRequest")) {
					sqlRequest = (String) map.get("sqlRequest");
				} else if (map.containsKey("username")) {
					username = (String) map.get("username");
				} else if (map.containsKey("password")) {
					password = (String) map.get("password");
				} else if (map.containsKey("attributeSupportedId")) {
					attributeSupportedId = (String) map
							.get("attributeSupportedId");
				} else if (map.containsKey("substituteValue")) {
					substituteValue = (String) map.get("substituteValue");
				}
			}
		}
		LOGGER.debug("Initialisation of the Attribute DB Connector");
		LOGGER.debug("URL = " + url);
		LOGGER.debug("Username = " + username);
		LOGGER.debug("Password = " + password);
		LOGGER.debug("DbName = " + dbName);
		LOGGER.debug("Driver = " + driver);
		LOGGER.debug("AttributeSupportedId = " + attributeSupportedId);
		LOGGER.debug("SqlRequest = " + sqlRequest);
		LOGGER.debug("SubstituteValue = " + substituteValue);
	}

	// always return true, since this is a feature we always support
	public boolean isDesignatorSupported() {
		return true;
	}

	// return a single identifier that shows support for environment attrs
	public Set getSupportedDesignatorTypes() {
		Set set = new HashSet();
		set.add(new Integer(AttributeDesignator.SUBJECT_TARGET));
		set.add(new Integer(AttributeDesignator.RESOURCE_TARGET));
		return set;
	}

	public EvaluationResult findAttribute(URI attributeType, URI attributeId,
			URI issuer, URI subjectCategory, EvaluationCtx context,
			int designatorType) {

		LOGGER.debug("Call of AttributeDBFinder to resolve attribute "
				+ attributeSupportedId);
		try {

			if (substituteValue == "urn:oasis:names:tc:xacml:1.0:resource:resource-id") {
				resourceCategoryUri = new URI(
						"urn:oasis:names:tc:xacml:1.0:resource-category:access-resource");
			} else {
				resourceCategoryUri = new URI(
						"urn:oasis:names:tc:xacml:1.0:subject-category:access-subject");
			}
			resourceId = new URI(substituteValue);
			resourceType = new URI("http://www.w3.org/2001/XMLSchema#string");
		} catch (URISyntaxException e1) {
			LOGGER.fatal("An exception occured : " + e1);
		}
		Set set = new HashSet();
		Connection con = null;

		// make sure this is an Subject or resource attribute
		if ((designatorType != AttributeDesignator.SUBJECT_TARGET)
				&& (designatorType != AttributeDesignator.RESOURCE_TARGET))
			return new EvaluationResult(
					BagAttribute.createEmptyBag(attributeType));

		// // make sure they're asking for our identifier
		if (!attributeId.toString().equals(attributeSupportedId))
			return new EvaluationResult(
					BagAttribute.createEmptyBag(attributeType));

		// make sure they're asking for an String return value
		if (!attributeType.toString().equals(StringAttribute.identifier))
			return new EvaluationResult(
					BagAttribute.createEmptyBag(attributeType));

		AttributeValue resource = context.getResourceAttribute(resourceType,
				resourceId, null).getAttributeValue();

		if (resource.isBag()) {
			BagAttribute attrBag = (BagAttribute) resource;
			Iterator it = attrBag.getValue().iterator();
			while (it.hasNext()) {
				resourceVal = (StringAttribute) it.next();
			}
			LOGGER.debug("Search attribute : " + attributeId
					+ " for resource : " + resourceVal.getValue());
		}
		// Test subject value
		if (resourceVal.getValue() == null) {
			LOGGER.error("Resource value is Null, Unable to retrieve "
					+ attributeId);
			return new EvaluationResult(
					BagAttribute.createEmptyBag(attributeType));
		}

		try {
			con = DBConnection();
			Statement stmt = con.createStatement();
			ResultSet result;
			String sql = parseSql(sqlRequest, resourceVal.encode(), "filter");
			sql = parseSql(sql, attributeSupportedId, "alias");
			LOGGER.debug("Final SQL request: " + sql);
			result = stmt.executeQuery(sql);
			while (result.next()) {
				set.add(new StringAttribute(result
						.getString(attributeSupportedId)));
			}
			con.close();
		} catch (SQLException e1) {
			LOGGER.fatal(e1.getLocalizedMessage());
		} catch (InstantiationException e2) {
			LOGGER.fatal(e2.getLocalizedMessage());
		}
		if (set.isEmpty()) {
			return new EvaluationResult(
					BagAttribute.createEmptyBag(attributeType));
		}

		return new EvaluationResult(new BagAttribute(attributeType, set));
	}

	// Method for DB Connection
	private Connection DBConnection() throws InstantiationException {
		String url = this.url;
		String dbName = this.dbName;
		String driver = this.driver;
		String username = this.username;
		String password = this.password;
		Connection con = null;

		try {
			Class.forName(driver).newInstance();
			con = DriverManager.getConnection(url + dbName, username, password);
		} catch (InstantiationException e) {
			throw (e);
		} catch (IllegalAccessException e1) {
			LOGGER.fatal(e1.getLocalizedMessage());
		} catch (ClassNotFoundException e2) {
			LOGGER.fatal(e2.getLocalizedMessage());
		} catch (SQLException e3) {
			LOGGER.fatal(e3.getLocalizedMessage());
		}

		return con;
	}

	private String parseSql(String request, String val, String arg) {
		return request.replace("$" + arg, "\"" + val + "\"");
	}
}
